# _*_ coding: utf-8 _*_
#
# hspylib-datasource v0.9.40
#
# Package: main.datasource.redis
"""Package initialization."""

__all__ = [
    'redis_configuration', 
    'redis_repository'
]
__version__ = '0.9.40'
