# _*_ coding: utf-8 _*_
#
# hspylib-datasource v0.9.40
#
# Package: main.datasource
"""Package initialization."""

__all__ = [
    'crud_entity', 
    'crud_repository', 
    'crud_service', 
    'db_configuration', 
    'db_repository', 
    'exception', 
    'firebase', 
    'identity', 
    'mysql', 
    'redis', 
    'sqlite'
]
__version__ = '0.9.40'
